/*
 * Copyright (c) by Shuu Yamaguchi <shuu@dotAster.com>
 *
 * $Id: call.c,v 1.5 2003/05/23 07:32:47 shuu Exp shuu $
 *
 * Can be freely distributed and used under the terms of the GNU GPL.
 */
#include	<stdio.h>
#include	<sys/types.h>
#include	<unistd.h>
#include	<string.h>
#include	<limits.h>
#include	<sys/wait.h>
#include	<stdlib.h>
#include	<signal.h>

#include	"aoi.h"

#ifndef	AOI_BIN
#define		AOI_BIN	"/etc/aoi/bin"
#endif

enum status {ST_NONE, ST_CALL};
enum child {CHLD_IGNORE, CHLD_WAIT};

static void
child_action(enum child chld)
{
	struct sigaction sa;
	static struct sigaction old_sa;
	
	if (chld == CHLD_IGNORE) {	/* must be called at first time */
		sa.sa_handler = SIG_IGN;
		sa.sa_flags = SA_NOCLDSTOP;
		sigemptyset(&sa.sa_mask);
		sigaction(SIGCHLD, &sa, &old_sa);
	} else {
		sigaction(SIGCHLD, &old_sa, NULL);
	}
}

static void
execute(char *device,char *fname)
{
	char *cmd[2];
	char path[PATH_MAX+1];
	char *envp[16];
	int i = 0;
	char *scratch,buf[128];

	syslog(LOG_LEVEL,"executing %s\n",fname);
#ifdef	DEBUG
	printf("[%s]:executing %s\n",device,fname);
#endif
	if (strlen(AOI_BIN) + 1 + strlen(fname) > PATH_MAX) {
		printf("path is too long\n");
		return;
	}
	strcpy(path,AOI_BIN);
	strcat(path,"/");
	strcat(path,fname);
	cmd[0] = fname;
	cmd[1] = NULL;
	scratch = buf;
	envp[i++] = "PATH=/etc/aoi/bin:/bin:/sbin:/usr/bin:/usr/sbin";
	envp[i++] = "HOME=/";
	envp[i++] = scratch;
	scratch += sprintf(scratch,"DEVICE=%s",device);
	envp[i] = NULL;

	switch(fork()) {
	case 0:		/* grandchild */
		execve(path,cmd,envp);
		break;
	default:	/* child */
		wait(NULL);	/* for serial execution */
		break;
	}
}

/*
 * Aoi(parent) ignores a child, but a child does Not ignore grandchildren.
 */
void
call(char *device,char **list)
{
	int i;
	enum status st = ST_NONE;

	child_action(CHLD_IGNORE);
	switch(fork()) {
	case 0:	/* child */
		for(i = 0;list[i] != NULL;i++) {
			if (st == ST_NONE) {
				if(strcmp(AOI_CALL,list[i]) == 0)
					st = ST_CALL;
			} else {
				child_action(CHLD_WAIT);
				execute(device,list[i]);
				st = ST_NONE;
			}
		}
		exit(0);
		break;
	default:	/* ignore child & Not create zombie */
		break;
	}
}
