/*
 * Copyright (c) by Shuu Yamaguchi <shuu@dotAster.com>
 *
 * $Id: match.c,v 1.4 2003/05/07 07:37:24 shuu Exp shuu $
 *
 * Can be freely distributed and used under the terms of the GNU GPL.
 */
#include	<stdio.h>
#include	<string.h>
#include	<sys/types.h>
#include	<regex.h>

#include	"aoi.h"

#define	LINE_MAX	2048

int
find_device(char ***table,char *name)
{
	int line;
	int col = 0;	/* for Warning */
	regex_t reg;

	for(line = 0;table[line] != NULL;line++) {
		for(col = 0;table[line][col] != NULL;col++) {
			if (strcmp(AOI_DEVICE,table[line][col]) == 0)
				goto loop_out;
		}
	}
loop_out:
	col++;
	if (regcomp(&reg,table[line][col],REG_NOSUB|REG_NEWLINE) != 0)
		return -1;
	if (regexec(&reg,name,0,NULL,0) == 0)
		return line;
	regfree(&reg);

	return -1;
}

static int
match(char *fname,char *string,enum aoi_match aoi_match)
{
	FILE *fp;
	char buf[LINE_MAX];
	int ret = AOI_NO;
	int match_flag = REG_NOSUB;
	regex_t reg;

	if (aoi_match == FAIL)
		return AOI_NO;
	if ((fp = fopen(fname,"r")) == NULL)
		return AOI_NO;
	if (aoi_match == STRING) {
		while(fgets(buf,LINE_MAX,fp) != NULL) {
			if (strstr(buf,string) != NULL) {
				ret = AOI_YES;
				break;
			}
		}
	} else {
		if (aoi_match == ERE)
			match_flag |= REG_EXTENDED;
		if (regcomp(&reg,string,match_flag) != 0)
			return AOI_NO;
		while(fgets(buf,LINE_MAX,fp) != NULL) {
			if (regexec(&reg,buf,0,NULL,0) == 0) {
				ret = AOI_YES;
				break;
			}
		}
		regfree(&reg);
	}
	fclose(fp);

	return ret;
}

enum status {ST_NONE,ST_FILE,ST_STRING,ST_BRE,ST_ERE};

int
match_conf(char **list)
{
	int i,ret = AOI_YES;
	char *fname = NULL;		/* for Warning */
	enum aoi_match flag = FAIL;	/* for Warning */
	enum status status = ST_NONE;

	for(i = 0;list[i] != NULL;i++) {
		if (status == ST_FILE) {
			fname = list[i];
			status = ST_NONE;
			continue;
		} else if ((status ==  ST_STRING) || 
			(status == ST_BRE) || (status == ST_ERE)) {
			if (match(fname,list[i],flag) == AOI_YES) {
				ret = AOI_YES;
#ifdef DEBUG
				printf("match %s:%s\n",fname,list[i]);
#endif
			} else {	/* if no match, immediately exit */
				ret = AOI_NO;
				break;
			}
			status = ST_NONE;
			continue;
		}
		if (strcmp(AOI_FILE,list[i]) == 0) {
			status = ST_FILE;
		} else if (strcmp(AOI_STRING,list[i]) == 0) {
			flag = STRING;
			status = ST_STRING;
		} else if (strcmp(AOI_BRE,list[i]) == 0) {
			flag = BRE;
			status = ST_BRE;
		} else if (strcmp(AOI_ERE,list[i]) == 0) {
			flag = ERE;
			status = ST_ERE;
		}
	}

	return ret;
}
