/*
 * Copyright (c) by Shuu Yamaguchi <shuu@dotAster.com>
 *
 * $Id: read_table.c,v 1.4 2003/05/07 07:49:49 shuu Exp shuu $
 *
 * Can be freely distributed and used under the terms of the GNU GPL.
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<ctype.h>
#include	<unistd.h>
#include	<sys/mman.h>

#include	"aoi.h"

#define	CHAR_SEP	'\n'
#define	CHAR_QUOTE	'\"'
#ifdef	ENABLE_ESCAPE
#define	CHAR_ESCAPE	'\\'
#endif	/* ENABLE_ESCAPE */


enum string_stat {STR_IN, STR_OUT, QUOTE_IN, QUOTE_OUT};

static char *
set_string(char *str_startp,char *str_endp)
{
	int len;
	char *ptr;

	len = str_endp - str_startp;
	ptr = (char *)malloc(len+1);
	if (ptr == NULL) 
		return NULL;
#ifdef	ENABLE_ESCAPE
	for(i = 0; str_startp < str_endp; str_startp++) {
		if (*str_startp == CHAR_ESCAPE) {
			continue;
		}
		ptr[i++] = *str_startp;
	}
#else
	memcpy(ptr,str_startp,len);
#endif	/* ENABLE_ESCAPE */
	ptr[len] = '\0';

	return ptr;
}

static char *
read_line(char *startp,char *endp,char **list,int *list_max)
{
	char *mp,*str_startp = NULL;	/* for Warning */
	int count = 0;
	enum string_stat status = STR_OUT;

	if (isspace(*startp))
		return NULL;
	for(mp = startp; mp <= endp; mp++) {
		if (status == STR_OUT) {
			if (isspace(*mp)) {
				if (*mp == CHAR_SEP) {
					mp++;
					break;
				}
				continue;
			} else {	/*  string start */
#ifdef	ENABLE_ESCAPE
				if (*mp == CHAR_ESCAPE) {
					if (mp < endp)
						mp++;
				} else 
#endif	/* ENABLE_ESCAPE */
				if (*mp == CHAR_QUOTE) {
					if (mp < endp)
						mp++;
					status = QUOTE_IN;
				} else {
					status = STR_IN;
				}
				if (count == *list_max) {
					printf("argument > max\n");
					return NULL;
				}
				str_startp = mp;
			}
		} else if (status == STR_IN) {
			if (isspace(*mp)) {
				list[count] = set_string(str_startp,mp);
				if (list[count] == NULL) 
					return NULL;
				count++;
				status = STR_OUT;
				if (*mp == CHAR_SEP) {
					mp++;
					break;
				}
			} else {
#ifdef	ENABLE_ESCAPE
				if (*mp == CHAR_ESCAPE) {
					if (mp < endp)
						mp++;
				}
#endif	/* ENABLE_ESCAPE */
				continue;
			}
		} else {	/* status == QUOTE_IN */
#ifdef	ENABLE_ESCAPE
			if (*mp == CHAR_ESCAPE) {
				if (mp < endp)
					mp++;
			} else 
#endif	/* ENABLE_ESCAPE */
			if (*mp == CHAR_QUOTE) {
				list[count] = set_string(str_startp,mp);
				if (list[count] == NULL) 
					return NULL;
				count++;
				status = STR_OUT;
			}
			continue;
		}
	}

#ifdef	DEBUG
	printf("count %d\n",count);
#endif
	*list_max = count;
	list[count] = NULL;

	return mp;
	
}

static char **
copy_line(char **line,int num)
{
	char **new_line;
	int i;

	new_line = (char **)malloc(sizeof(char **) * (num+1));
	if (new_line == NULL)
		return NULL;
	for(i = 0;i < num;i++) {
		new_line[i] = line[i];
	}
	new_line[i] = NULL;

	return new_line;
	
}

static void
read_lines(char *startp,char *endp,char ***table)
{
	char *mp;
#ifdef	DEBUG
	int i,count = 0;
#endif
	int list_max;
	int line_num;
	char *tmp_list[LIST_MAX+1];	/* last item is kept as NULL */

	for(mp = startp,line_num = 0;mp < endp;line_num++)  {
		list_max = LIST_MAX;
		mp = read_line(mp,endp,tmp_list,&list_max);
		if (mp == NULL) {
			printf("read_line return NULL\n");
		}
		table[line_num] = copy_line(tmp_list,list_max);
#ifdef	DEBUG
		for(i = 0,count++;table[line_num][i] != NULL; i++) {
			printf("%d line[%d]=\"%s\"\n",count,i,table[line_num][i]);
		}
#endif
	}
}

static int
table_line(char *startp,char *endp)
{
	int count;

	for(count = 0;startp <= endp;startp++) {
		if (*startp == '\n')
			count++;
	}

	return count;
}

char ***
read_conf(char *fname)
{
	int fd;
	struct stat st;
	char *mp;
	int line;
	char ***table;

	if ((fd = open(fname,O_RDONLY)) == -1) 
		return NULL;
	if (fstat(fd,&st) == -1)
		return NULL;
	if ((mp = mmap(0,st.st_size,PROT_READ,MAP_SHARED,fd,0)) == MAP_FAILED)
		return NULL;
	close(fd);
	line = table_line(mp,mp+st.st_size);
#ifdef	DEBUG
	printf("table line %d\n",line);
#endif
	table = (char ***)malloc(sizeof(char ***) * (line+1));
	if (table == NULL)
		return NULL;
	table[line] = NULL;
	read_lines(mp,mp+st.st_size,table);
	munmap(mp,st.st_size);

	return table;
}

